@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2013 David Robillard <http://drobilla.net>
+  Copyright 2011-2014 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -25,7 +25,7 @@
 #define ZIX_INLINE
 #include "zix/digest.c"
 #include "zix/hash.c"
-#include "zix/tree.c"
+#include "zix/btree.c"
 
 #include "sord_config.h"
 #include "sord_internal.h"
@@ -83,11 +83,13 @@ typedef enum {
 	GPOS,  ///< Graph,  Predicate, Object,    Subject
 } SordOrder;
 
+#ifdef SORD_DEBUG_SEARCH
 /** String name of each ordering (array indexed by SordOrder) */
 static const char* const order_names[NUM_ORDERS] = {
 	"spo",  "sop",  "ops",  "osp",  "pso",  "pos",
 	"gspo", "gsop", "gops", "gosp", "gpso", "gpos"
 };
+#endif
 
 /**
    Quads of indices for each order, from most to least significant
@@ -116,9 +118,10 @@ struct SordModelImpl {
 	/** Index for each possible triple ordering (may or may not exist).
 	 * Each index is a tree of SordQuad with the appropriate ordering.
 	 */
-	ZixTree* indices[NUM_ORDERS];
+	ZixBTree* indices[NUM_ORDERS];
 
 	size_t n_quads;
+	size_t n_iters;
 };
 
 /** Mode for searching or iteration */
@@ -133,9 +136,9 @@ typedef enum {
 /** Iterator over some range of a store */
 struct SordIterImpl {
 	const SordModel* sord;               ///< Model being iterated over
-	ZixTreeIter*     cur;                ///< Current DB cursor
+	ZixBTreeIter*    cur;                ///< Current DB cursor
 	SordQuad         pat;                ///< Pattern (in ordering order)
-	int              ordering[TUP_LEN];  ///< Store ordering
+	SordOrder        order;              ///< Store order (which index)
 	SearchMode       mode;               ///< Iteration mode
 	int              n_prefix;           ///< Prefix for RANGE and FILTER_RANGE
 	bool             end;                ///< True iff reached end
@@ -199,9 +202,9 @@ sord_world_new(void)
 }
 
 static void
-free_node_entry(const void* value, void* user_data)
+free_node_entry(void* value, void* user_data)
 {
-	const SordNode* node = (const SordNode*)value;
+	SordNode* node = (SordNode*)value;
 	if (node->node.type == SERD_LITERAL) {
 		sord_node_free((SordWorld*)user_data, node->meta.lit.datatype);
 	}
@@ -316,36 +319,36 @@ static inline bool
 sord_iter_forward(SordIter* iter)
 {
 	if (!iter->skip_graphs) {
-		iter->cur = zix_tree_iter_next(iter->cur);
-		return zix_tree_iter_is_end(iter->cur);
+		zix_btree_iter_increment(iter->cur);
+		return zix_btree_iter_is_end(iter->cur);
 	}
 
-	SordNode** key = (SordNode**)zix_tree_get(iter->cur);
+	SordNode**     key     = (SordNode**)zix_btree_get(iter->cur);
 	const SordQuad initial = { key[0], key[1], key[2], key[3] };
-	while (true) {
-		iter->cur = zix_tree_iter_next(iter->cur);
-		if (zix_tree_iter_is_end(iter->cur))
-			return true;
-
-		key = (SordNode**)zix_tree_get(iter->cur);
+	zix_btree_iter_increment(iter->cur);
+	while (!zix_btree_iter_is_end(iter->cur)) {
+		key = (SordNode**)zix_btree_get(iter->cur);
 		for (int i = 0; i < 3; ++i)
 			if (key[i] != initial[i])
 				return false;
+
+		zix_btree_iter_increment(iter->cur);
 	}
-	assert(false);
+
+	return true;
 }
 
 /**
-   Seek forward as necessary until @a iter points at a match.
+   Seek forward as necessary until `iter` points at a match.
    @return true iff iterator reached end of valid range.
 */
 static inline bool
 sord_iter_seek_match(SordIter* iter)
 {
 	for (iter->end = true;
-	     !zix_tree_iter_is_end(iter->cur);
+	     !zix_btree_iter_is_end(iter->cur);
 	     sord_iter_forward(iter)) {
-		const SordNode** const key = (const SordNode**)zix_tree_get(iter->cur);
+		const SordNode** const key = (const SordNode**)zix_btree_get(iter->cur);
 		if (sord_quad_match_inline(key, iter->pat))
 			return (iter->end = false);
 	}
@@ -353,24 +356,23 @@ sord_iter_seek_match(SordIter* iter)
 }
 
 /**
-   Seek forward as necessary until @a iter points at a match, or the prefix
+   Seek forward as necessary until `iter` points at a match, or the prefix
    no longer matches iter->pat.
    @return true iff iterator reached end of valid range.
 */
 static inline bool
 sord_iter_seek_match_range(SordIter* iter)
 {
-	if (iter->end)
-		return true;
+	assert(!iter->end);
 
 	do {
-		const SordNode** key = (const SordNode**)zix_tree_get(iter->cur);
+		const SordNode** key = (const SordNode**)zix_btree_get(iter->cur);
 
 		if (sord_quad_match_inline(key, iter->pat))
 			return false;  // Found match
 
 		for (int i = 0; i < iter->n_prefix; ++i) {
-			const int idx = iter->ordering[i];
+			const int idx = orderings[iter->order][i];
 			if (!sord_id_match(key[idx], iter->pat[idx])) {
 				iter->end = true;  // Reached end of valid range
 				return true;
@@ -382,21 +384,19 @@ sord_iter_seek_match_range(SordIter* iter)
 }
 
 static SordIter*
-sord_iter_new(const SordModel* sord, ZixTreeIter* cur, const SordQuad pat,
+sord_iter_new(const SordModel* sord, ZixBTreeIter* cur, const SordQuad pat,
               SordOrder order, SearchMode mode, int n_prefix)
 {
-	const int* ordering = orderings[order];
-
 	SordIter* iter = (SordIter*)malloc(sizeof(SordIter));
 	iter->sord        = sord;
 	iter->cur         = cur;
+	iter->order       = order;
 	iter->mode        = mode;
 	iter->n_prefix    = n_prefix;
 	iter->end         = false;
 	iter->skip_graphs = order < GSPO;
 	for (int i = 0; i < TUP_LEN; ++i) {
-		iter->pat[i]      = pat[i];
-		iter->ordering[i] = ordering[i];
+		iter->pat[i] = pat[i];
 	}
 
 	switch (iter->mode) {
@@ -404,7 +404,7 @@ sord_iter_new(const SordModel* sord, ZixTreeIter* cur, const SordQuad pat,
 	case SINGLE:
 	case RANGE:
 		assert(
-			sord_quad_match_inline((const SordNode**)zix_tree_get(iter->cur),
+			sord_quad_match_inline((const SordNode**)zix_btree_get(iter->cur),
 			                       iter->pat));
 		break;
 	case FILTER_RANGE:
@@ -422,6 +422,8 @@ sord_iter_new(const SordModel* sord, ZixTreeIter* cur, const SordQuad pat,
 	              (void*)iter, TUP_FMT_ARGS(pat), TUP_FMT_ARGS(value),
 	              iter->end, iter->skip_graphs);
 #endif
+
+	++((SordModel*)sord)->n_iters;
 	return iter;
 }
 
@@ -434,7 +436,7 @@ sord_iter_get_model(SordIter* iter)
 void
 sord_iter_get(const SordIter* iter, SordQuad id)
 {
-	SordNode** key = (SordNode**)zix_tree_get(iter->cur);
+	SordNode** key = (SordNode**)zix_btree_get(iter->cur);
 	for (int i = 0; i < TUP_LEN; ++i) {
 		id[i] = key[i];
 	}
@@ -443,17 +445,17 @@ sord_iter_get(const SordIter* iter, SordQuad id)
 const SordNode*
 sord_iter_get_node(const SordIter* iter, SordQuadIndex index)
 {
-	return iter ? ((SordNode**)zix_tree_get(iter->cur))[index] : NULL;
+	return iter ? ((SordNode**)zix_btree_get(iter->cur))[index] : NULL;
 }
 
-bool
-sord_iter_next(SordIter* iter)
+static bool
+sord_iter_scan_next(SordIter* iter)
 {
-	if (iter->end)
+	if (iter->end) {
 		return true;
+	}
 
 	const SordNode** key;
-	iter->end = sord_iter_forward(iter);
 	if (!iter->end) {
 		switch (iter->mode) {
 		case ALL:
@@ -466,10 +468,10 @@ sord_iter_next(SordIter* iter)
 		case RANGE:
 			SORD_ITER_LOG("%p range next\n", (void*)iter);
 			// At the end if the MSNs no longer match
-			key = (const SordNode**)zix_tree_get(iter->cur);
+			key = (const SordNode**)zix_btree_get(iter->cur);
 			assert(key);
 			for (int i = 0; i < iter->n_prefix; ++i) {
-				const int idx = iter->ordering[i];
+				const int idx = orderings[iter->order][i];
 				if (!sord_id_match(key[idx], iter->pat[idx])) {
 					iter->end = true;
 					SORD_ITER_LOG("%p reached non-match end\n", (void*)iter);
@@ -504,6 +506,17 @@ sord_iter_next(SordIter* iter)
 	}
 }
 
+bool
+sord_iter_next(SordIter* iter)
+{
+	if (iter->end) {
+		return true;
+	}
+
+	iter->end = sord_iter_forward(iter);
+	return sord_iter_scan_next(iter);
+}
+
 bool
 sord_iter_end(const SordIter* iter)
 {
@@ -515,13 +528,15 @@ sord_iter_free(SordIter* iter)
 {
 	SORD_ITER_LOG("%p Free\n", (void*)iter);
 	if (iter) {
+		--((SordModel*)iter->sord)->n_iters;
+		zix_btree_iter_free(iter->cur);
 		free(iter);
 	}
 }
 
 /**
-   Return true iff @a sord has an index for @a order.
-   If @a graphs is true, @a order will be modified to be the
+   Return true iff `sord` has an index for `order`.
+   If `graphs` is true, `order` will be modified to be the
    corresponding order with a G prepended (so G will be the MSN).
 */
 static inline bool
@@ -540,7 +555,7 @@ sord_has_index(SordModel* sord, SordOrder* order, int* n_prefix, bool graphs)
    @param pat Pattern in standard (S P O G) order
    @param mode Set to the (best) iteration mode for iterating over results
    @param n_prefix Set to the length of the range prefix
-   (for @a mode == RANGE and @a mode == FILTER_RANGE)
+   (for `mode` == RANGE and `mode` == FILTER_RANGE)
 */
 static inline SordOrder
 sord_best_index(SordModel*     sord,
@@ -570,8 +585,10 @@ sord_best_index(SordModel*     sord,
 	*n_prefix = 0;
 	switch (sig) {
 	case 0x000:
-		*mode = ALL;
-		return graph_search ? DEFAULT_GRAPH_ORDER : DEFAULT_ORDER;
+		assert(graph_search);
+		*mode     = RANGE;
+		*n_prefix = 1;
+		return DEFAULT_GRAPH_ORDER;
 	case 0x111:
 		*mode = SINGLE;
 		return graph_search ? DEFAULT_GRAPH_ORDER : DEFAULT_ORDER;
@@ -597,7 +614,7 @@ sord_best_index(SordModel*     sord,
 	switch (sig) {
 		PAT_CASE(0x011, FILTER_RANGE, OSP, PSO, 1);
 		PAT_CASE(0x101, FILTER_RANGE, SPO, OPS, 1);
-		PAT_CASE(0x110, FILTER_RANGE, SOP, POS, 1);
+		// SPO is always present, so 0x110 is never reached here
 	default: break;
 	}
 
@@ -625,17 +642,18 @@ sord_new(SordWorld* world, unsigned indices, bool graphs)
 	SordModel* sord = (SordModel*)malloc(sizeof(struct SordModelImpl));
 	sord->world   = world;
 	sord->n_quads = 0;
+	sord->n_iters = 0;
 
 	for (unsigned i = 0; i < (NUM_ORDERS / 2); ++i) {
 		const int* const ordering   = orderings[i];
 		const int* const g_ordering = orderings[i + (NUM_ORDERS / 2)];
 
 		if (indices & (1 << i)) {
-			sord->indices[i] = zix_tree_new(
-				false, sord_quad_compare, (void*)ordering, NULL);
+			sord->indices[i] = zix_btree_new(
+				sord_quad_compare, (void*)ordering, NULL);
 			if (graphs) {
-				sord->indices[i + (NUM_ORDERS / 2)] = zix_tree_new(
-					false, sord_quad_compare, (void*)g_ordering, NULL);
+				sord->indices[i + (NUM_ORDERS / 2)] = zix_btree_new(
+					sord_quad_compare, (void*)g_ordering, NULL);
 			} else {
 				sord->indices[i + (NUM_ORDERS / 2)] = NULL;
 			}
@@ -646,12 +664,12 @@ sord_new(SordWorld* world, unsigned indices, bool graphs)
 	}
 
 	if (!sord->indices[DEFAULT_ORDER]) {
-		sord->indices[DEFAULT_ORDER] = zix_tree_new(
-			false, sord_quad_compare, (void*)orderings[DEFAULT_ORDER], NULL);
+		sord->indices[DEFAULT_ORDER] = zix_btree_new(
+			sord_quad_compare, (void*)orderings[DEFAULT_ORDER], NULL);
 	}
 	if (graphs && !sord->indices[DEFAULT_GRAPH_ORDER]) {
-		sord->indices[DEFAULT_GRAPH_ORDER] = zix_tree_new(
-			false, sord_quad_compare, (void*)orderings[DEFAULT_GRAPH_ORDER], NULL);
+		sord->indices[DEFAULT_GRAPH_ORDER] = zix_btree_new(
+			sord_quad_compare, (void*)orderings[DEFAULT_GRAPH_ORDER], NULL);
 	}
 
 	return sord;
@@ -721,16 +739,16 @@ sord_free(SordModel* sord)
 	sord_iter_free(i);
 
 	// Free quads
-	for (ZixTreeIter* t = zix_tree_begin(sord->indices[DEFAULT_ORDER]);
-	     !zix_tree_iter_is_end(t);
-	     t = zix_tree_iter_next(t)) {
-		free(zix_tree_get(t));
+	ZixBTreeIter* t = zix_btree_begin(sord->indices[DEFAULT_ORDER]);
+	for (; !zix_btree_iter_is_end(t); zix_btree_iter_increment(t)) {
+		free(zix_btree_get(t));
 	}
+	zix_btree_iter_free(t);
 
 	// Free indices
 	for (unsigned o = 0; o < NUM_ORDERS; ++o)
 		if (sord->indices[o])
-			zix_tree_free(sord->indices[o]);
+			zix_btree_free(sord->indices[o]);
 
 	free(sord);
 }
@@ -759,46 +777,12 @@ sord_begin(const SordModel* sord)
 	if (sord_num_quads(sord) == 0) {
 		return NULL;
 	} else {
-		ZixTreeIter* cur = zix_tree_begin(sord->indices[DEFAULT_ORDER]);
-		SordQuad pat = { 0, 0, 0, 0 };
+		ZixBTreeIter* cur = zix_btree_begin(sord->indices[DEFAULT_ORDER]);
+		SordQuad      pat = { 0, 0, 0, 0 };
 		return sord_iter_new(sord, cur, pat, DEFAULT_ORDER, ALL, 0);
 	}
 }
 
-static inline ZixTreeIter*
-index_search(ZixTree* db, const SordQuad search_key)
-{
-	ZixTreeIter* iter = NULL;
-	zix_tree_find(db, (const void*)search_key, &iter);
-	return iter;
-}
-
-static inline ZixTreeIter*
-index_lower_bound(ZixTree* db, const SordQuad search_key)
-{
-	ZixTreeIter* iter = NULL;
-	zix_tree_find(db, (const void*)search_key, &iter);
-	if (!iter) {
-		return NULL;
-	}
-
-	ZixTreeIter* prev = NULL;
-	while ((prev = zix_tree_iter_prev(iter))) {
-		if (!prev) {
-			return iter;
-		}
-
-		const SordNode** const key = (const SordNode**)zix_tree_get(prev);
-		if (!sord_quad_match_inline(key, search_key)) {
-			return iter;
-		}
-
-		iter = prev;
-	}
-
-	return iter;
-}
-
 SordIter*
 sord_find(SordModel* sord, const SordQuad pat)
 {
@@ -809,24 +793,25 @@ sord_find(SordModel* sord, const SordQuad pat)
 	int             n_prefix;
 	const SordOrder index_order = sord_best_index(sord, pat, &mode, &n_prefix);
 
-	SORD_FIND_LOG("Find " TUP_FMT "  index=%s  mode=%d"
-	              "  n_prefix=%d ordering=%d%d%d%d\n",
-	              TUP_FMT_ARGS(pat), order_names[index_order], mode, n_prefix,
-	              ordering[0], ordering[1], ordering[2], ordering[3]);
+	SORD_FIND_LOG("Find " TUP_FMT "  index=%s  mode=%d  n_prefix=%d\n",
+	              TUP_FMT_ARGS(pat), order_names[index_order], mode, n_prefix);
 
 	if (pat[0] && pat[1] && pat[2] && pat[3])
 		mode = SINGLE;  // No duplicate quads (Sord is a set)
 
-	ZixTree* const     db  = sord->indices[index_order];
-	ZixTreeIter* const cur = index_lower_bound(db, pat);
-	if (zix_tree_iter_is_end(cur)) {
+	ZixBTree* const db  = sord->indices[index_order];
+	ZixBTreeIter*   cur = NULL;
+	zix_btree_lower_bound(db, pat, &cur);
+	if (zix_btree_iter_is_end(cur)) {
 		SORD_FIND_LOG("No match found\n");
+		zix_btree_iter_free(cur);
 		return NULL;
 	}
-	const SordNode** const key = (const SordNode**)zix_tree_get(cur);
+	const SordNode** const key = (const SordNode**)zix_btree_get(cur);
 	if (!key || ( (mode == RANGE || mode == SINGLE)
 	              && !sord_quad_match_inline(pat, key) )) {
 		SORD_FIND_LOG("No match found\n");
+		zix_btree_iter_free(cur);
 		return NULL;
 	}
 
@@ -917,16 +902,14 @@ SordNodeType
 sord_node_get_type(const SordNode* node)
 {
 	switch (node->node.type) {
-	case SERD_BLANK:
-		return SORD_BLANK;
-	case SERD_LITERAL:
-		return SORD_LITERAL;
 	case SERD_URI:
 		return SORD_URI;
+	case SERD_BLANK:
+		return SORD_BLANK;
 	default:
-		fprintf(stderr, "error: invalid node type\n");
-		return (SordNodeType)0;
+		return SORD_LITERAL;
 	}
+	SORD_UNREACHABLE();
 }
 
 const uint8_t*
@@ -1074,7 +1057,7 @@ sord_new_literal_counted(SordWorld*     world,
 	SordNode key = {
 		{ str, n_bytes, n_chars, flags, SERD_LITERAL }, 1, { { 0 } }
 	};
-	key.meta.lit.datatype = datatype;
+	key.meta.lit.datatype = sord_node_copy(datatype);
 	memset(key.meta.lit.lang, 0, sizeof(key.meta.lit.lang));
 	if (lang) {
 		strncpy(key.meta.lit.lang, lang, sizeof(key.meta.lit.lang));
@@ -1177,10 +1160,9 @@ sord_node_free(SordWorld* world, SordNode* node)
 {
 	if (!node) {
 		return;
-	}
-
-	assert(node->refs > 0);
-	if (--node->refs == 0) {
+	} else if (node->refs == 0) {
+		error(world, SERD_ERR_BAD_ARG, "attempt to free garbage node\n");
+	} else if (--node->refs == 0) {
 		sord_node_free_internal(world, node);
 	}
 }
@@ -1198,7 +1180,7 @@ sord_node_copy(const SordNode* node)
 static inline bool
 sord_add_to_index(SordModel* sord, const SordNode** tup, SordOrder order)
 {
-	return !zix_tree_insert(sord->indices[order], tup, NULL);
+	return !zix_btree_insert(sord->indices[order], tup);
 }
 
 bool
@@ -1209,13 +1191,15 @@ sord_add(SordModel* sord, const SordQuad tup)
 		error(sord->world, SERD_ERR_BAD_ARG,
 		      "attempt to add quad with NULL field\n");
 		return false;
+	} else if (sord->n_iters > 0) {
+		error(sord->world, SERD_ERR_BAD_ARG, "added tuple during iteration\n");
 	}
 
 	const SordNode** quad = (const SordNode**)malloc(sizeof(SordQuad));
 	memcpy(quad, tup, sizeof(SordQuad));
 
 	for (unsigned i = 0; i < NUM_ORDERS; ++i) {
-		if (sord->indices[i]) {
+		if (sord->indices[i] && (i < GSPO || tup[3])) {
 			if (!sord_add_to_index(sord, quad, (SordOrder)i)) {
 				assert(i == 0);  // Assuming index coherency
 				free(quad);
@@ -1235,17 +1219,14 @@ void
 sord_remove(SordModel* sord, const SordQuad tup)
 {
 	SORD_WRITE_LOG("Remove " TUP_FMT "\n", TUP_FMT_ARGS(tup));
+	if (sord->n_iters > 0) {
+		error(sord->world, SERD_ERR_BAD_ARG, "remove with iterator\n");
+	}
 
-	SordNode** quad = NULL;
+	SordNode* quad = NULL;
 	for (unsigned i = 0; i < NUM_ORDERS; ++i) {
-		if (sord->indices[i]) {
-			ZixTreeIter* const cur = index_search(sord->indices[i], tup);
-			if (!zix_tree_iter_is_end(cur)) {
-				if (!quad) {
-					quad = (SordNode**)zix_tree_get(cur);
-				}
-				zix_tree_remove(sord->indices[i], cur);
-			} else {
+		if (sord->indices[i] && (i < GSPO || tup[3])) {
+			if (zix_btree_remove(sord->indices[i], tup, (void**)&quad, NULL)) {
 				assert(i == 0);  // Assuming index coherency
 				return;  // Quad not found, do nothing
 			}
@@ -1259,3 +1240,37 @@ sord_remove(SordModel* sord, const SordQuad tup)
 
 	--sord->n_quads;
 }
+
+SerdStatus
+sord_erase(SordModel* sord, SordIter* iter)
+{
+	if (sord->n_iters > 1) {
+		error(sord->world, SERD_ERR_BAD_ARG, "erased with many iterators\n");
+		return SERD_ERR_BAD_ARG;
+	}
+
+	SordQuad tup;
+	sord_iter_get(iter, tup);
+
+	SORD_WRITE_LOG("Remove " TUP_FMT "\n", TUP_FMT_ARGS(tup));
+
+	SordNode* quad = NULL;
+	for (unsigned i = 0; i < NUM_ORDERS; ++i) {
+		if (sord->indices[i]) {
+			if (zix_btree_remove(sord->indices[i], tup, (void**)&quad,
+			                     i == iter->order ? &iter->cur : NULL)) {
+				return (i == 0) ? SERD_ERR_NOT_FOUND : SERD_ERR_INTERNAL;
+			}
+		}
+	}
+	iter->end = zix_btree_iter_is_end(iter->cur);
+	sord_iter_scan_next(iter);
+
+	free(quad);
+
+	for (int i = 0; i < TUP_LEN; ++i)
+		sord_drop_quad_ref(sord, tup[i], (SordQuadIndex)i);
+
+	--sord->n_quads;
+	return SERD_SUCCESS;
+}
