@@ -1,5 +1,5 @@
 /*
-  Copyright 2011 David Robillard <http://drobilla.net>
+  Copyright 2011-2013 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -22,9 +22,10 @@
 #include "sord/sord.h"
 
 static const int      DIGITS        = 3;
-static const int      MAX_NUM       = 999;
 static const unsigned n_objects_per = 2;
 
+static int n_expected_errors = 0;
+
 typedef struct {
 	SordQuad query;
 	int      expected_num_results;
@@ -73,7 +74,7 @@ generate(SordWorld* world,
 		}
 
 		for (unsigned j = 0; j < n_objects_per; ++j) {
-			SordQuad tup = { ids[0], ids[1], ids[2 + j] };
+			SordQuad tup = { ids[0], ids[1], ids[2 + j], graph };
 			if (!sord_add(sord, tup)) {
 				return test_fail("Fail: Failed to add quad\n");
 			}
@@ -87,11 +88,11 @@ generate(SordWorld* world,
 	// Add some literals
 
 	// (98 4 "hello") and (98 4 "hello"^^<5>)
-	SordQuad tup = { 0, 0, 0, 0};
+	SordQuad tup = { 0, 0, 0, 0 };
 	tup[0] = uri(world, 98);
 	tup[1] = uri(world, 4);
 	tup[2] = sord_new_literal(world, 0, USTR("hello"), NULL);
-	tup[3] = 0;
+	tup[3] = graph;
 	sord_add(sord, tup);
 	sord_node_free(world, (SordNode*)tup[2]);
 	tup[2] = sord_new_literal(world, uri(world, 5), USTR("hello"), NULL);
@@ -103,7 +104,7 @@ generate(SordWorld* world,
 	tup[0] = uri(world, 96);
 	tup[1] = uri(world, 4);
 	tup[2] = sord_new_literal(world, uri(world, 4), USTR("hello"), NULL);
-	tup[3] = 0;
+	tup[3] = graph;
 	sord_add(sord, tup);
 	sord_node_free(world, (SordNode*)tup[2]);
 	tup[2] = sord_new_literal(world, uri(world, 5), USTR("hello"), NULL);
@@ -115,7 +116,7 @@ generate(SordWorld* world,
 	tup[0] = uri(world, 94);
 	tup[1] = uri(world, 5);
 	tup[2] = sord_new_literal(world, 0, USTR("hello"), NULL);
-	tup[3] = 0;
+	tup[3] = graph;
 	sord_add(sord, tup);
 	sord_node_free(world, (SordNode*)tup[2]);
 	tup[2] = sord_new_literal(world, NULL, USTR("hello"), "en-gb");
@@ -127,7 +128,7 @@ generate(SordWorld* world,
 	tup[0] = uri(world, 92);
 	tup[1] = uri(world, 6);
 	tup[2] = sord_new_literal(world, 0, USTR("hello"), "en-us");
-	tup[3] = 0;
+	tup[3] = graph;
 	sord_add(sord, tup);
 	sord_node_free(world, (SordNode*)tup[2]);
 	tup[2] = sord_new_literal(world, NULL, USTR("hello"), "en-gb");
@@ -205,10 +206,11 @@ test_read(SordWorld* world, SordModel* sord, SordNode* g,
 	SordNode* gb_hello    = sord_new_literal(world, NULL, s, "en-gb");
 	SordNode* us_hello    = sord_new_literal(world, NULL, s, "en-us");
 
-#define NUM_PATTERNS 17
+#define NUM_PATTERNS 18
 
 	QueryTest patterns[NUM_PATTERNS] = {
 		{ { 0, 0, 0 }, (n_quads * n_objects_per) + 12 },
+		{ { uri(world, 1), 0, 0 }, 2 },
 		{ { uri(world, 9), uri(world, 9), uri(world, 9) }, 0 },
 		{ { uri(world, 1), uri(world, 2), uri(world, 4) }, 1 },
 		{ { uri(world, 3), uri(world, 4), uri(world, 0) }, 2 },
@@ -238,6 +240,22 @@ test_read(SordWorld* world, SordModel* sord, SordNode* g,
 		                 TUP_FMT_ARGS(nomatch));
 	}
 
+	if (sord_get(sord, NULL, NULL, uri(world, 3), g)) {
+		return test_fail("Fail: Get *,*,3 succeeded\n");
+	} else if (!sord_node_equals(
+		           sord_get(sord, uri(world, 1), uri(world, 2), NULL, g),
+		           uri(world, 3))) {
+		return test_fail("Fail: Get 1,2,* != 3\n");
+	} else if (!sord_node_equals(
+		           sord_get(sord, uri(world, 1), NULL, uri(world, 3), g),
+		           uri(world, 2))) {
+		return test_fail("Fail: Get 1,*,3 != 2\n");
+	} else if (!sord_node_equals(
+		           sord_get(sord, NULL, uri(world, 2), uri(world, 3), g),
+		           uri(world, 1))) {
+		return test_fail("Fail: Get *,2,3 != 1\n");
+	}
+
 	for (unsigned i = 0; i < NUM_PATTERNS; ++i) {
 		QueryTest test = patterns[i];
 		SordQuad  pat = { test.query[0], test.query[1], test.query[2], g };
@@ -337,6 +355,23 @@ test_read(SordWorld* world, SordModel* sord, SordNode* g,
 	return ret;
 }
 
+static SerdStatus
+unexpected_error(void* handle, const SerdError* error)
+{
+	fprintf(stderr, "unexpected error: ");
+	vfprintf(stderr, error->fmt, *error->args);
+	return SERD_SUCCESS;
+}
+
+static SerdStatus
+expected_error(void* handle, const SerdError* error)
+{
+	fprintf(stderr, "expected error: ");
+	vfprintf(stderr, error->fmt, *error->args);
+	++n_expected_errors;
+	return SERD_SUCCESS;
+}
+
 int
 main(int argc, char** argv)
 {
@@ -346,6 +381,58 @@ main(int argc, char** argv)
 
 	SordWorld* world = sord_world_new();
 
+
+	// Attempt to create invalid URI
+	fprintf(stderr, "expected ");
+	SordNode* bad_uri = sord_new_uri(world, USTR("noscheme"));
+	if (bad_uri) {
+		return test_fail("Successfully created invalid URI \"noscheme\"\n");
+	}
+	sord_node_free(world, bad_uri);
+
+	sord_world_set_error_sink(world, expected_error, NULL);
+
+	// Attempt to create invalid CURIE
+	SerdNode  base   = serd_node_from_string(SERD_URI, USTR("http://example.org/"));
+	SerdEnv*  env    = serd_env_new(&base);
+	SerdNode  sbadns = serd_node_from_string(SERD_CURIE, USTR("badns:"));
+	SordNode* badns  = sord_node_from_serd_node(world, env, &sbadns, NULL, NULL);
+	if (badns) {
+		return test_fail("Successfully created CURIE with bad namespace\n");
+	}
+	sord_node_free(world, badns);
+	serd_env_free(env);
+
+	// Attempt to create NULL node
+	SordNode* nil_node = sord_node_from_serd_node(
+		world, NULL, &SERD_NODE_NULL, NULL, NULL);
+	if (nil_node) {
+		return test_fail("Successfully created NULL node\n");
+	}
+	sord_node_free(world, nil_node);
+
+	// Attempt to double-free a node
+	SordNode* garbage = sord_new_uri(world, USTR("urn:garbage"));
+	sord_node_free(world, garbage);
+	sord_world_set_error_sink(world, expected_error, NULL);
+	sord_node_free(world, garbage);
+	sord_world_set_error_sink(world, unexpected_error, NULL);
+	if (n_expected_errors != 2) {
+		return test_fail("Successfully freed node twice\n");
+	}
+
+	sord_world_set_error_sink(world, unexpected_error, NULL);
+
+	// Check node flags are set properly
+	SordNode* with_newline = sord_new_literal(world, NULL, USTR("a\nb"), NULL);
+	if (!(sord_node_get_flags(with_newline) & SERD_HAS_NEWLINE)) {
+		return test_fail("Newline flag not set\n");
+	}
+	SordNode* with_quote = sord_new_literal(world, NULL, USTR("a\"b"), NULL);
+	if (!(sord_node_get_flags(with_quote) & SERD_HAS_QUOTE)) {
+		return test_fail("Quote flag not set\n");
+	}
+
 	// Create with minimal indexing
 	SordModel* sord = sord_new(world, SORD_SPO, false);
 	generate(world, sord, n_quads, NULL);
@@ -357,6 +444,7 @@ main(int argc, char** argv)
 	}
 
 	// Check adding tuples with NULL fields fails
+	sord_world_set_error_sink(world, expected_error, NULL);
 	const size_t initial_num_quads = sord_num_quads(sord);
 	SordQuad tup = { 0, 0, 0, 0};
 	if (sord_add(sord, tup)) {
@@ -376,6 +464,28 @@ main(int argc, char** argv)
 		                 sord_num_quads(sord), initial_num_quads);
 	}
 
+	// Check adding tuples with an active iterator fails
+	SordIter* iter = sord_begin(sord);
+	tup[2] = uri(world, 3);
+	if (sord_add(sord, tup)) {
+		return test_fail("Added tuple with active iterator\n");
+	}
+
+	// Check removing tuples with several active iterator fails
+	SordIter* iter2 = sord_begin(sord);
+	if (!sord_erase(sord, iter)) {
+		return test_fail("Erased tuple with several active iterators\n");
+	}
+	n_expected_errors = 0;
+	sord_remove(sord, tup);
+	if (n_expected_errors != 1) {
+		return test_fail("Removed tuple with several active iterators\n");
+	}
+	sord_iter_free(iter);
+	sord_iter_free(iter2);
+
+	sord_world_set_error_sink(world, unexpected_error, NULL);
+
 	// Check interning merges equivalent values
 	SordNode* uri_id   = sord_new_uri(world, USTR("http://example.org"));
 	SordNode* blank_id = sord_new_blank(world, USTR("testblank"));
@@ -502,7 +612,7 @@ main(int argc, char** argv)
 	}
 
 	// Test removing
-	sord = sord_new(world, SORD_SPO, false);
+	sord = sord_new(world, SORD_SPO, true);
 	tup[0] = uri(world, 1);
 	tup[1] = uri(world, 2);
 	tup[2] = sord_new_literal(world, 0, USTR("hello"), NULL);
@@ -517,15 +627,92 @@ main(int argc, char** argv)
 	sord_add(sord, tup);
 	sord_remove(sord, tup);
 	if (sord_num_quads(sord) != 1) {
-		fprintf(stderr, "Removed failed (%zu quads, expected 1)\n",
+		fprintf(stderr, "Remove failed (%zu quads, expected 1)\n",
 		        sord_num_quads(sord));
 		goto fail;
 	}
 
+	iter = sord_find(sord, tup);
+	if (!sord_iter_end(iter)) {
+		fprintf(stderr, "Found removed tuple\n");
+		goto fail;
+	}
+	sord_iter_free(iter);
+
+	// Load a couple graphs
+	SordNode* graph42 = uri(world, 42);
+	SordNode* graph43 = uri(world, 43);
+	generate(world, sord, 1, graph42);
+	generate(world, sord, 1, graph43);
+
+	// Remove one graph via iterator
+	SerdStatus st;
+	iter = sord_search(sord, NULL, NULL, NULL, graph43);
+	while (!sord_iter_end(iter)) {
+		if ((st = sord_erase(sord, iter))) {
+			fprintf(stderr, "Remove by iterator failed (%s)\n",
+			        serd_strerror(st));
+			goto fail;
+		}
+	}
+	sord_iter_free(iter);
+
+	// Ensure only the other graph is left
+	SordQuad quad;
+	SordQuad pat = { 0, 0, 0, graph42 };
+	for (iter = sord_begin(sord); !sord_iter_end(iter); sord_iter_next(iter)) {
+		sord_iter_get(iter, quad);
+		if (!sord_quad_match(quad, pat)) {
+			fprintf(stderr, "Graph removal via iteration failed\n");
+			goto fail;
+		}
+	}
+	sord_iter_free(iter);
+
+	// Load file into two separate graphs
 	sord_free(sord);
+	sord = sord_new(world, SORD_SPO, true);
+	env  = serd_env_new(&base);
+	SordNode*   graph1 = sord_new_uri(world, USTR("http://example.org/graph1"));
+	SordNode*   graph2 = sord_new_uri(world, USTR("http://example.org/graph2"));
+	SerdReader* reader = sord_new_reader(sord, env, SERD_TURTLE, graph1);
+	if ((st = serd_reader_read_string(reader, USTR("<s> <p> <o> .")))) {
+		fprintf(stderr, "Failed to read string (%s)\n", serd_strerror(st));
+		goto fail;
+	}
+	serd_reader_free(reader);
+	reader = sord_new_reader(sord, env, SERD_TURTLE, graph2);
+	if ((st = serd_reader_read_string(reader, USTR("<s> <p> <o> .")))) {
+		fprintf(stderr, "Failed to re-read string (%s)\n", serd_strerror(st));
+		goto fail;
+	}
 
+	// Ensure we only see triple once
+	size_t n_triples = 0;
+	for (iter = sord_begin(sord); !sord_iter_end(iter); sord_iter_next(iter)) {
+		fprintf(stderr, "%s %s %s %s\n",
+		        sord_node_get_string(sord_iter_get_node(iter, SORD_SUBJECT)),
+		        sord_node_get_string(sord_iter_get_node(iter, SORD_PREDICATE)),
+		        sord_node_get_string(sord_iter_get_node(iter, SORD_OBJECT)),
+		        sord_node_get_string(sord_iter_get_node(iter, SORD_GRAPH)));
+
+		++n_triples;
+	}
+	if (n_triples != 1) {
+		fprintf(stderr, "Found duplicate triple\n");
+		goto fail;
+	}
+
+	// Test SPO iteration on an SOP indexed store
+	sord_free(sord);
+	sord = sord_new(world, SORD_SOP, false);
+	generate(world, sord, 1, graph42);
+	for (iter = sord_begin(sord); !sord_iter_end(iter); sord_iter_next(iter)) {
+		++n_triples;
+	}
+
+	sord_free(sord);
 	sord_world_free(world);
-
 	return EXIT_SUCCESS;
 
 fail:
