@@ -1,5 +1,5 @@
 /*
-  Copyright 2011-2012 David Robillard <http://drobilla.net>
+  Copyright 2011-2013 David Robillard <http://drobilla.net>
 
   Permission to use, copy, modify, and/or distribute this software for any
   purpose with or without fee is hereby granted, provided that the above
@@ -28,6 +28,9 @@
 #include "sord/sord.h"
 #include "sord_config.h"
 
+#define SORDI_ERROR(msg)       fprintf(stderr, "sordi: " msg);
+#define SORDI_ERRORF(fmt, ...) fprintf(stderr, "sordi: " fmt, __VA_ARGS__);
+
 typedef struct {
 	SerdWriter* writer;
 	SerdEnv*    env;
@@ -40,7 +43,7 @@ static int
 print_version(void)
 {
 	printf("sordi " SORD_VERSION " <http://drobilla.net/software/sord>\n");
-	printf("Copyright 2011-2012 David Robillard <http://drobilla.net>.\n"
+	printf("Copyright 2011-2013 David Robillard <http://drobilla.net>.\n"
 	       "License: <http://www.opensource.org/licenses/isc>\n"
 	       "This is free software; you are free to change and redistribute it."
 	       "\nThere is NO WARRANTY, to the extent permitted by law.\n");
@@ -51,6 +54,7 @@ static int
 print_usage(const char* name, bool error)
 {
 	FILE* const os = error ? stderr : stdout;
+	fprintf(os, "%s", error ? "\n" : "");
 	fprintf(os, "Usage: %s [OPTION]... INPUT [BASE_URI]\n", name);
 	fprintf(os, "Load and re-serialise RDF data.\n");
 	fprintf(os, "Use - for INPUT to read from standard input.\n\n");
@@ -70,7 +74,7 @@ set_syntax(SerdSyntax* syntax, const char* name)
 	} else if (!strcmp(name, "ntriples")) {
 		*syntax = SERD_NTRIPLES;
 	} else {
-		fprintf(stderr, "Unknown input format `%s'\n", name);
+		SORDI_ERRORF("unknown syntax `%s'\n", name);
 		return false;
 	}
 	return true;
@@ -117,29 +121,29 @@ main(int argc, char** argv)
 			break;
 		} else if (argv[a][1] == 'i') {
 			if (++a == argc) {
-				fprintf(stderr, "Missing value for -i\n");
-				return 1;
+				SORDI_ERROR("option requires an argument -- 'i'\n\n");
+				return print_usage(argv[0], true);
 			}
 			if (!set_syntax(&input_syntax, argv[a])) {
-				return 1;
+				return print_usage(argv[0], true);
 			}
 		} else if (argv[a][1] == 'o') {
 			if (++a == argc) {
-				fprintf(stderr, "Missing value for -o\n");
-				return 1;
+				SORDI_ERROR("option requires an argument -- 'o'\n\n");
+				return print_usage(argv[0], true);
 			}
 			if (!set_syntax(&output_syntax, argv[a])) {
-				return 1;
+				return print_usage(argv[0], true);
 			}
 		} else {
-			fprintf(stderr, "Unknown option `%s'\n", argv[a]);
+			SORDI_ERRORF("invalid option -- '%s'\n", argv[a] + 1);
 			return print_usage(argv[0], true);
 		}
 	}
 
 	if (a == argc) {
-		fprintf(stderr, "Missing input\n");
-		return 1;
+		SORDI_ERROR("missing input\n");
+		return print_usage(argv[0], true);
 	}
 
 	const uint8_t* input   = (const uint8_t*)argv[a++];
@@ -159,13 +163,14 @@ main(int argc, char** argv)
 	if (a < argc) {  // Base URI given on command line
 		base_uri_node = serd_node_new_uri_from_string(
 			(const uint8_t*)argv[a], NULL, &base_uri);
-	} else if (from_file) {  // Use input file URI
+	} else if (in_fd != stdin && from_file) {  // Use input file URI
 		base_uri_node = serd_node_new_file_uri(in_path, NULL, &base_uri, false);
 	}
 
+	free(in_path);
 	if (!base_uri_node.buf) {
-		fprintf(stderr, "Missing base URI\n");
-		return 1;
+		SORDI_ERROR("missing base URI\n");
+		return print_usage(argv[0], true);
 	}
 
 	SordWorld*  world  = sord_world_new();
